@@ -1,5 +1,5 @@
 /*
-Copyright (C) 2004-2011 John E. Davis
+Copyright (C) 2004-2021,2022 John E. Davis
 
 This file is part of the S-Lang Library.
 
@@ -25,6 +25,10 @@ USA.
 # include <slang.h>
 #endif
 
+#ifdef __cplusplus
+extern "C" {
+#endif
+
 /* This is a temporary hack until lynx is fixed to not include this file. */
 #ifndef LYCURSES_H
 
@@ -71,7 +75,7 @@ extern int SLcurses_wnoutrefresh (SLcurses_Window_Type *);
 extern int SLcurses_wclrtoeol (SLcurses_Window_Type *);
 
 extern int SLcurses_wmove (SLcurses_Window_Type *, unsigned int, unsigned int);
-extern int SLcurses_waddch (SLcurses_Window_Type *, SLtt_Char_Type);
+extern int SLcurses_waddch (SLcurses_Window_Type *, SLcurses_Char_Type);
 extern int SLcurses_waddnstr (SLcurses_Window_Type *, char *, int);
 
 #define waddnstr		SLcurses_waddnstr
@@ -105,7 +109,7 @@ extern int SLcurses_scanw (char *, ...);
 #define scanw SLcurses_scanw
 #endif
 
-extern SLcurses_Window_Type *SLcurses_Stdscr;
+SL_EXTERN SLcurses_Window_Type *SLcurses_Stdscr;
 #define WINDOW SLcurses_Window_Type
 #define stdscr SLcurses_Stdscr
 
@@ -142,9 +146,9 @@ extern int SLcurses_nil (void);
 extern int SLcurses_wgetch (SLcurses_Window_Type *);
 extern int SLcurses_getch (void);
 
-extern int SLcurses_wattrset (SLcurses_Window_Type *, SLtt_Char_Type);
-extern int SLcurses_wattron (SLcurses_Window_Type *, SLtt_Char_Type);
-extern int SLcurses_wattroff (SLcurses_Window_Type *, SLtt_Char_Type);
+extern int SLcurses_wattrset (SLcurses_Window_Type *, SLcurses_Char_Type);
+extern int SLcurses_wattron (SLcurses_Window_Type *, SLcurses_Char_Type);
+extern int SLcurses_wattroff (SLcurses_Window_Type *, SLcurses_Char_Type);
 #define attrset(x) SLcurses_wattrset(stdscr, (x))
 #define attron(x) SLcurses_wattron(stdscr, (x))
 #define attroff(x) SLcurses_wattroff(stdscr, (x))
@@ -204,7 +208,7 @@ extern int SLcurses_raw (void);
 
 extern int SLcurses_endwin (void);
 #define endwin SLcurses_endwin
-extern int SLcurses_Is_Endwin;
+SL_EXTERN int SLcurses_Is_Endwin;
 #define isendwin() SLcurses_Is_Endwin
 
 #define keypad(w,x) ((w)->use_keypad = (x))
@@ -242,7 +246,7 @@ extern int SLcurses_Is_Endwin;
    (w)->scroll_max=(w)->nrows, \
    wscrl((w), -1))
 
-extern SLtt_Char_Type SLcurses_Acs_Map [128];
+SL_EXTERN SLcurses_Char_Type SLcurses_Acs_Map [128];
 #define acs_map SLcurses_Acs_Map
 
 #define ACS_ULCORNER (acs_map[SLSMG_ULCORN_CHAR])
@@ -286,6 +290,7 @@ extern SLtt_Char_Type SLcurses_Acs_Map [128];
 #define A_UNDERLINE	0x40000000UL	/* was 0x4000 */
 #define A_ALTCHARSET	0x80000000UL	/* was 0x8000 */
 #define A_BLINK		0x00000000UL
+#define A_ITALIC	0x00000000UL
 #define A_DIM		0x00000000UL
 #define A_PROTECT	0x00000000UL
 #define A_INVIS		0x00000000UL
@@ -299,12 +304,19 @@ extern SLtt_Char_Type SLcurses_Acs_Map [128];
 #define COLOR_CYAN	SLSMG_COLOR_CYAN
 #define COLOR_WHITE	SLSMG_COLOR_LGRAY
 
-extern int SLcurses_Num_Colors;
+SL_EXTERN int SLcurses_Num_Colors;
 #define COLORS		SLcurses_Num_Colors
 #define COLOR_PAIRS	(SLcurses_Num_Colors*SLcurses_Num_Colors)
 
-#define init_pair(_x,_f,_b) \
- SLtt_set_color_object((_x), ((_f) == (_b) ? 0x0700 : ((_f) | ((_b) << 8)) << 8))
+#if SLANG_VERSION >= 30000
+# define init_pair(_x,_f,_b) \
+   SLtt_set_color_fgbg ((_x), \
+			(((_f) == (_b)) ? 7 : (_f)), \
+			(((_f) == (_b)) ? 0 : (_b)))
+#else
+# define init_pair(_x,_f,_b) \
+   SLtt_set_color_object((_x), ((_f) == (_b) ? 0x0700 : ((_f) | ((_b) << 8)) << 8))
+#endif
 
 #define scrollok(a,b) ((a)->scroll_ok = (b))
 #define getyx(a,y,x)  (y=(a)->_cury, x=(a)->_curx)
@@ -330,7 +342,7 @@ extern int SLcurses_wdelch (SLcurses_Window_Type *);
 extern int SLcurses_winsch (SLcurses_Window_Type *, int);
 #define winsch SLcurses_winsch
 
-extern int SLcurses_Esc_Delay;/* ESC expire time in milliseconds (ncurses compatible) */
+SL_EXTERN int SLcurses_Esc_Delay;/* ESC expire time in milliseconds (ncurses compatible) */
 #define ESCDELAY SLcurses_Esc_Delay
 
 extern int SLcurses_clearok (SLcurses_Window_Type *, int);
@@ -358,7 +370,9 @@ extern int SLcurses_clearok (SLcurses_Window_Type *, int);
 /* These have no place in C */
 #define TRUE 1
 #define FALSE 0
-#define bool int
+#if !defined(_STDBOOL)  /* MSVC v18+ */
+# define bool int
+#endif
 
 /* Lynx compatability */
 #else
@@ -379,3 +393,7 @@ extern int SLcurses_clearok (SLcurses_Window_Type *, int);
 #define endwin SLsmg_reset_smg(),SLang_reset_tty
 
 #endif
+
+#ifdef __cplusplus
+}
+#endif
